#include "php_gtk.h"
#if HAVE_PHP_GTK

#include "gtkcombobutton.h"
#include "ext/gtk+/php_gtk+.h"
#define GTK_COMBO_BUTTON    GTK_COMBOBUTTON

void php_combobutton_register_constants(int module_number TSRMLS_DC)
{


}

PHP_GTK_EXPORT_CE(gtk_combobutton_ce);

PHP_FUNCTION(gtk_combobutton_new_with_label)
{
	char *label;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "s", &label))
		return;

	PHP_GTK_SEPARATE_RETURN(return_value, php_gtk_new((GtkObject *)gtk_combobutton_new_with_label(label)));
	return;
}


PHP_FUNCTION(gtk_combobutton_new_with_menu)
{
	zval *menu;
	zend_bool changing;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ob", &menu, gtk_widget_ce, &changing))
		return;

	PHP_GTK_SEPARATE_RETURN(return_value, php_gtk_new((GtkObject *)gtk_combobutton_new_with_menu(GTK_WIDGET(PHP_GTK_GET(menu)), (gboolean)changing)));
	return;
}

PHP_GTK_EXPORT_CE(combobutton_ce);

static function_entry php_combobutton_functions[] = {
	{"ton_new_with_label",	PHP_FN(gtk_combobutton_new_with_label),	NULL},
	{"ton_new_with_menu",	PHP_FN(gtk_combobutton_new_with_menu),	NULL},
	{NULL, NULL, NULL}
};


/* object GtkCombobutton  */
PHP_FUNCTION(gtk_combobutton_new)
{
	zval *arg = NULL;
	GtkObject *wrapped_obj;
	zend_bool changing = 0;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|Vb", &arg, &changing)) {
		php_gtk_invalidate(this_ptr);
		return;
	}

	if (arg) {
		if (Z_TYPE_P(arg) == IS_STRING) {
			wrapped_obj = (GtkObject *)gtk_combobutton_new_with_label(estrdup(Z_STRVAL_P(arg)));
		} else if (Z_TYPE_P(arg) == IS_OBJECT && php_gtk_check_class(arg, gtk_menu_ce)) {
			wrapped_obj = (GtkObject *)gtk_combobutton_new_with_menu(GTK_WIDGET(PHP_GTK_GET(arg)), changing);
		} else {
			php_error(E_WARNING, "%s(): expecting argument 1 to be string or GtkMenu object",
					  get_active_function_name(TSRMLS_C));
			php_gtk_invalidate(this_ptr);
			return;
		}
	} else {
		wrapped_obj = (GtkObject *)gtk_combobutton_new();
	}
	if (!wrapped_obj) {
		php_error(E_WARNING, "%s(): could not create GtkComboButton object",
				  get_active_function_name(TSRMLS_C));
		php_gtk_invalidate(this_ptr);
		return;
	}

	php_gtk_object_init(wrapped_obj, this_ptr);
}

PHP_FUNCTION(gtk_combobutton_get_type)
{
	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	RETURN_LONG(gtk_combobutton_get_type());
}


PHP_FUNCTION(gtk_combobutton_get_menu)
{
	NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	PHP_GTK_SEPARATE_RETURN(return_value, php_gtk_new((GtkObject *)gtk_combobutton_get_menu(GTK_COMBOBUTTON(PHP_GTK_GET(this_ptr)))));
	return;
}


PHP_FUNCTION(gtk_combobutton_set_menu)
{
	zval *combomenu;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &combomenu, gtk_widget_ce))
		return;

	gtk_combobutton_set_menu(GTK_COMBOBUTTON(PHP_GTK_GET(this_ptr)), GTK_WIDGET(PHP_GTK_GET(combomenu)));
	RETURN_NULL();
}


PHP_FUNCTION(gtk_combobutton_get_arrow_direction)
{
	long php_retval;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	php_retval = gtk_combobutton_get_arrow_direction(GTK_COMBOBUTTON(PHP_GTK_GET(this_ptr)));
	RETURN_LONG(php_retval);
}


PHP_FUNCTION(gtk_combobutton_set_arrow_direction)
{
	GtkArrowType arrowtype;
	zval *php_arrowtype = NULL;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_arrowtype))
		return;

	if (php_arrowtype && !php_gtk_get_enum_value(GTK_TYPE_ARROW_TYPE, php_arrowtype, (gint *)&arrowtype)) {
		return;
	}

	gtk_combobutton_set_arrow_direction(GTK_COMBOBUTTON(PHP_GTK_GET(this_ptr)), arrowtype);
	RETURN_NULL();
}


static function_entry php_gtk_combobutton_functions[] = {
	{"GtkCombobutton",	PHP_FN(gtk_combobutton_new),	NULL},
	{"gtkcombobutton",	PHP_FN(gtk_combobutton_new),	NULL},
	{"get_type",	PHP_FN(gtk_combobutton_get_type),	NULL},
	{"get_menu",	PHP_FN(gtk_combobutton_get_menu),	NULL},
	{"set_menu",	PHP_FN(gtk_combobutton_set_menu),	NULL},
	{"get_arrow_direction",	PHP_FN(gtk_combobutton_get_arrow_direction),	NULL},
	{"set_arrow_direction",	PHP_FN(gtk_combobutton_set_arrow_direction),	NULL},
	{NULL, NULL, NULL}
};


static void gtk_combobutton_get_property(zval *return_value, zval *object, zend_llist_element **element, int *result)
{
	char *prop_name = Z_STRVAL(((zend_overloaded_element *)(*element)->data)->element);

	*result = SUCCESS;

	if (!strcmp(prop_name, "x_offset")) {
		long php_retval;

	 	php_retval = GTK_COMBOBUTTON(PHP_GTK_GET(object))->x_offset;
		RETURN_LONG(php_retval);
	} else if (!strcmp(prop_name, "y_offset")) {
		long php_retval;

	 	php_retval = GTK_COMBOBUTTON(PHP_GTK_GET(object))->y_offset;
		RETURN_LONG(php_retval);
	} else if (!strcmp(prop_name, "arrowdir")) {
		long php_retval;

	 	php_retval = GTK_COMBOBUTTON(PHP_GTK_GET(object))->arrowdir;
		RETURN_LONG(php_retval);
	}

	*result = FAILURE;
}


void php_combobutton_register_classes(void)
{
	zend_class_entry ce;
	TSRMLS_FETCH();

	INIT_OVERLOADED_CLASS_ENTRY(ce, "ComboButton", php_combobutton_functions, NULL, NULL, php_gtk_set_property);
	combobutton_ce = zend_register_internal_class_ex(&ce, NULL, NULL TSRMLS_CC);

	INIT_OVERLOADED_CLASS_ENTRY(ce, "GtkComboButton", php_gtk_combobutton_functions, NULL, php_gtk_get_property, php_gtk_set_property);
	gtk_combobutton_ce = zend_register_internal_class_ex(&ce, gtk_button_ce, NULL TSRMLS_CC);
	g_hash_table_insert(php_gtk_class_hash, g_strdup("GtkComboButton"), gtk_combobutton_ce);
	php_gtk_register_prop_getter(gtk_combobutton_ce, gtk_combobutton_get_property);


}

#endif /* HAVE_PHP_GTK */
